package eu.asvachin.audiorouter;

import java.util.logging.Level;
import java.util.logging.Logger;
import javax.sound.sampled.*;

/**
 * reads/writes from/to a target/source DataLine.
 *
 * @author Miles Asvachin
 */
public abstract class EndPoint<T extends DataLine> implements Runnable
{
	protected final T line;
	protected final byte[] data;
	protected final Buffer buffer;
	
	private Thread thread;
	//Have to use this flag instead of relying on Thread.isInterrupted because SourceDataLine.write eats the InterruptedException.
	private volatile boolean run;
	
	protected EndPoint(final Class<T> lineClass, final Mixer.Info newMixerInfo, final AudioFormat newFormat, final Buffer newBuffer) throws LineUnavailableException
    {
		if(newMixerInfo == null || newFormat == null || newBuffer == null)
		{
			throw new NullPointerException();
		}

		line = getDataLine(lineClass, newMixerInfo, newFormat, newBuffer.getSize());
		data = new byte[newBuffer.getSize()];
		buffer = newBuffer;
		
		if(line == null)
		{
			throw new LineUnavailableException();
		}
		
		run = true;
    }
	
	/**
	 * Asks the thread running this EndPoint to stop, and waits until it does, or a timeout occurs.
	 */
	public synchronized final void stop()
	{
		if(run)
		{
			run = false;
			buffer.reset();
			thread.interrupt();

			try
			{
				thread.join(5000L);
			}
			catch(InterruptedException ex)
			{
			}

			line.stop();
			line.close();

			if(thread.isAlive())
			{
				Logger.getLogger(getClass().getName()).log(Level.SEVERE, "Thread still alive after attempting to stop.");
				//@todo Can this be cleaned up better? The next attempt to connect will probably fail.
			}
		}
	}
	
	@Override
	public final void run()
	{
		thread = Thread.currentThread();
		
		while(run)
		{
			transfer();
		}
	}
	
	/**
	 * Do the actual work in here.
	 */
	protected abstract void transfer();
	
	/**
	 * Looks for a DataLine on the specified Mixer
	 *
	 * @param mixerInfo Mixer.Info of the Mixer to look for a line on.
	 * @param format the AudioFormat that the line must support.
	 * @param bufferSamples the buffer size the line must support, in bytes.
	 * @returns a DataLine guaranteed to support the given parameters, or null if one could not be found.
	 */
	@SuppressWarnings("unchecked") //Ignore the warning, assume that getLine works.
	public static <T extends DataLine> T getDataLine(final Class<T> lineClass, final Mixer.Info mixerInfo, final AudioFormat format, final int bufferSize)
	{
		final Line.Info lineInfo = new DataLine.Info(lineClass, format, bufferSize);
		T line = null;
		
		try
		{
			line = (T)AudioSystem.getMixer(mixerInfo).getLine(lineInfo);
		}
		catch(LineUnavailableException | IllegalArgumentException | SecurityException ex)
		{
		}
		
		return line;
	}
}
