package eu.asvachin.audiorouter.gui;

import javax.sound.sampled.*;
import javax.swing.DefaultListCellRenderer;
import javax.swing.JComboBox;
import javax.swing.JSpinner;
import javax.swing.SpinnerNumberModel;

/**
 *
 * @author Miles Asvachin
 */
public abstract class AudioControlCreator
{
	private static final Integer[] BUFFER_SIZES =
	{
		128,
		256,
		512,
		1024,
		2048,
		4096,
		8192,
		16384
	};
	
	private static final Integer[] SAMPLE_RATES =
	{
		8000,
		11025,
		16000,
		22050,
		24000,
		32000,
		44100,
		48000,
		88200,
		96000,
		192000
	};
	
	private static final Integer[] RESOLUTIONS =
	{
		8,
		16,
		24,
		32
	};

	/**
	 * Finds all potential audio inputs.
	 *
	 * @param defaultMixerInfo the item to be selected by default
	 * @return a JComboBox filled with Mixer.Infos that have at least one TargetDataLine.
	 */
	public static JComboBox<Mixer.Info> createInputCombo(final Mixer.Info defaultMixerInfo)
	{
		return createMixerInfoCombo(new Line.Info(TargetDataLine.class), defaultMixerInfo);
	}
	
	/**
	 * Finds all potential audio outputs.
	 *
	 * @param defaultMixerInfo the item to be selected by default
	 * @return a JComboBox filled with Mixer.Infos that have at least one SourceDataLine.
	 */
	public static JComboBox<Mixer.Info> createOutputCombo(final Mixer.Info defaultMixerInfo)
	{
		return createMixerInfoCombo(new Line.Info(SourceDataLine.class), defaultMixerInfo);
	}
	
	private static JComboBox<Mixer.Info> createMixerInfoCombo(final Line.Info desiredLine, final Mixer.Info defaultMixerInfo)
	{
		final MixerInfoListCellRenderer<Mixer.Info> listCellRenderer = new MixerInfoListCellRenderer<>();
		final JComboBox<Mixer.Info> combo = new JComboBox<>();
		combo.setRenderer(listCellRenderer);
		
		for(Mixer.Info mixerInfo : AudioSystem.getMixerInfo())
		{
			if(AudioSystem.getMixer(mixerInfo).isLineSupported(desiredLine))
			{
				combo.addItem(mixerInfo);
				
				if(mixerInfo == defaultMixerInfo)
					combo.setSelectedItem(mixerInfo);
			}
		}
		
		return combo;
	}
	
	/**
	 * Creates a combo box containing possible buffer sizes in bytes.
	 *
	 * @param defaultBufferSize the item to be selected by default
	 * @return a JComboBox filled with buffer sizes.
	 */
	public static JComboBox<Integer> createBufferCombo(final int defaultBufferSize)
	{
		return createFormatCombo(BUFFER_SIZES, defaultBufferSize);
	}
	
	/**
	 * Creates a combo box containing possible sample rates in Hz
	 *
	 * @param defaultSampleRate the item to be selected by default
	 * @return a JComboBox filled with sample rates.
	 */
	public static JComboBox<Integer> createSampleRateCombo(final int defaultSampleRate)
	{
		return createFormatCombo(SAMPLE_RATES, defaultSampleRate);
	}
	
	/**
	 * Creates a combo box containing possible bit depths.
	 *
	 * @param defaultResolution the item to be selected by default
	 * @return a JComboBox filled with bit depths.
	 */
	public static JComboBox<Integer> createResolutionCombo(final int defaultResolution)
	{
		return createFormatCombo(RESOLUTIONS, defaultResolution);
	}
	
	private static JComboBox<Integer> createFormatCombo(final Integer[] array, final int defaultSelection)
	{
		//Right-aligned list cell renderer.
		final DefaultListCellRenderer lRenderer = new DefaultListCellRenderer();
		lRenderer.setHorizontalAlignment(DefaultListCellRenderer.RIGHT);

		final JComboBox<Integer> combo = new JComboBox<>(array);
		combo.setRenderer(lRenderer);

		for(Integer item : array)
		{
			if(item.equals(defaultSelection))
			{
				combo.setSelectedItem(item);
				break;
			}
		}

		return combo;
	}
	
	/**
	 * Creates a spinner suitable for selecting number of desired audio channels.
	 *
	 * @param defaultChannels the default number of channels
	 * @return a JSpinner.
	 */
	public static JSpinner createChannelSpinner(final int defaultChannels)
	{
		return new JSpinner(new SpinnerNumberModel(defaultChannels, 1, 256, 1));
	}
}
