package eu.asvachin.audiorouter.gui;

import eu.asvachin.audiorouter.Main.Connector;
import eu.asvachin.toolbox.properties.Properties;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.Mixer;
import javax.swing.*;

/**
 *
 * @author Miles Asvachin
 */
public class AudioRouterGui
{
	private final JComponent contentPane;
	private JComboBox<Mixer.Info> inputCombo, outputCombo;
	private JComboBox<Integer> sampleCombo, resolutionCombo;
	private JComboBox<Integer> bufferCombo;
	private JButton connectButton;
	private JSpinner channelsSpinner;
	
	private final Connector connector;
	
	public AudioRouterGui(final Connector newConnector, final Mixer.Info selectedInput, final Mixer.Info selectedOutput, final AudioFormat selectedFormat, final int selectedBufferSize)
	{
		connector = newConnector;

		try
		{
			UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
		}
		catch(ClassNotFoundException | InstantiationException | IllegalAccessException | UnsupportedLookAndFeelException ex)
		{
			Logger.getLogger(getClass().getName()).log(Level.WARNING, null, ex);
		}
		
		contentPane = new JPanel();
		
		final Box devicePanel = new Box(BoxLayout.Y_AXIS);
		fillDevicePanel(devicePanel, selectedInput, selectedOutput);
		//Also add "connect" button to this panel.
		devicePanel.add(Box.createVerticalStrut(20));
		connectButton = new JButton("Connect");
		connectButton.setAlignmentX(Component.CENTER_ALIGNMENT);
		connectButton.setActionCommand("CONNECT");
		connectButton.setToolTipText("Connect input to output");
		connectButton.addActionListener(new ActionListener()
		{
			@Override
			public void actionPerformed(ActionEvent e)
			{
				if("CONNECT".equals(e.getActionCommand()))
				{
					if(connect())
					{
						disableControls();
						saveProperties();
					}
				}
				else
				{
					disconnect();
					enableControls();
				}
			}
		});
		devicePanel.add(connectButton);
		contentPane.add(devicePanel);

		final Box formatPanel = new Box(BoxLayout.Y_AXIS);
		fillFormatPanel(formatPanel, selectedFormat, selectedBufferSize);
		contentPane.add(formatPanel);
		
		JFrame.setDefaultLookAndFeelDecorated(true);
		final JFrame frame = new JFrame("Audio Router");
		frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		//frame.setIconImage(new ImageIcon(getClass().getResource("Icon32.png"), "Icon32").getImage());
		frame.setContentPane(contentPane);
		frame.setResizable(false);
		frame.pack();
		frame.setVisible(true);
	}
	
	/**
	 * Fills a panel with lists of input and output devices.
	 *
	 * @param container container to place device lists in
	 */
	private void fillDevicePanel(final Container container, final Mixer.Info selectedInput, final Mixer.Info selectedOutput)
	{
		//Find all inputs (target lines)
		JPanel panel = new JPanel(new FlowLayout(FlowLayout.TRAILING));
		JLabel label = new JLabel("Input");
		panel.add(label);
		inputCombo = AudioControlCreator.createInputCombo(selectedInput);
		panel.add(inputCombo);
		container.add(panel);

		//Find all outputs (source lines)
		panel = new JPanel(new FlowLayout(FlowLayout.TRAILING));
		label = new JLabel("Output");
		panel.add(label);
		outputCombo = AudioControlCreator.createOutputCombo(selectedOutput);
		panel.add(outputCombo);
		container.add(panel);

		//Equalize dimensions
		int comboWidth = (int) Math.max(inputCombo.getPreferredSize().getWidth(), outputCombo.getPreferredSize().getWidth());
		int comboHeight = (int) Math.max(inputCombo.getPreferredSize().getHeight(), outputCombo.getPreferredSize().getHeight());
		Dimension dimension = new Dimension(comboWidth, comboHeight);
		outputCombo.setPreferredSize(dimension);
		inputCombo.setPreferredSize(dimension);
	}

	/**
	 * Fills a panel with audio format options (sampling rate etc).
	 *
	 * @param container container to place audio format options in
	 */
	private void fillFormatPanel(final Container container, final AudioFormat selectedFormat, final int selectedBufferSize)
	{
		//Right-aligned list cell renderer to align text in combo boxes.
		DefaultListCellRenderer lRenderer = new DefaultListCellRenderer();
		lRenderer.setHorizontalAlignment(DefaultListCellRenderer.RIGHT);

		//Buffer size combo box
		JPanel panel = new JPanel(new FlowLayout(FlowLayout.TRAILING));
		JLabel label = new JLabel("Buffer size (samples)");
		panel.add(label);
		bufferCombo = AudioControlCreator.createBufferCombo(selectedBufferSize);
		panel.add(bufferCombo);
		container.add(panel);

		//Sample rate combo box
		panel = new JPanel(new FlowLayout(FlowLayout.TRAILING));
		label = new JLabel("Sample rate (Hz)");
		panel.add(label);
		sampleCombo = AudioControlCreator.createSampleRateCombo((int)selectedFormat.getSampleRate());
		panel.add(sampleCombo);
		container.add(panel);

		//Resolution combo box
		panel = new JPanel(new FlowLayout(FlowLayout.TRAILING));
		label = new JLabel("Resolution (bits)");
		panel.add(label);
		resolutionCombo = AudioControlCreator.createResolutionCombo(selectedFormat.getSampleSizeInBits());
		panel.add(resolutionCombo);
		container.add(panel);

		//Channels spinner
		panel = new JPanel(new FlowLayout(FlowLayout.TRAILING));
		label = new JLabel("Channels");
		panel.add(label);
		channelsSpinner = AudioControlCreator.createChannelSpinner(selectedFormat.getChannels());
		panel.add(channelsSpinner);
		container.add(panel);

		//Equalize dimensions
		Dimension dimension = new Dimension(sampleCombo.getPreferredSize());
		bufferCombo.setPreferredSize(dimension);
		resolutionCombo.setPreferredSize(dimension);
		channelsSpinner.setPreferredSize(dimension);
	}
	
	private void disableControls()
	{
		inputCombo.setEnabled(false);
		outputCombo.setEnabled(false);
		sampleCombo.setEnabled(false);
		resolutionCombo.setEnabled(false);
		bufferCombo.setEnabled(false);
		channelsSpinner.setEnabled(false);

		connectButton.setActionCommand("DISCONNECT");
		connectButton.setText("Disconnect");
		connectButton.setToolTipText("Disconnect input from output");
	}

	private void enableControls()
	{
		inputCombo.setEnabled(true);
		outputCombo.setEnabled(true);
		sampleCombo.setEnabled(true);
		resolutionCombo.setEnabled(true);
		bufferCombo.setEnabled(true);
		channelsSpinner.setEnabled(true);

		connectButton.setActionCommand("CONNECT");
		connectButton.setText("Connect");
		connectButton.setToolTipText("Connect input to output");
	}

	/**
	 * Connects input and output together.
	 */
	private boolean connect()
	{
		final int sampleRate = (Integer)sampleCombo.getSelectedItem();
		final int resolution = (Integer)resolutionCombo.getSelectedItem();
		final int channels = (Integer)channelsSpinner.getValue();
		//Buffer size in bytes = samples * bytes per sample * channels
		final int bufferSamples = (Integer)bufferCombo.getSelectedItem();
		final int bufferSize = bufferSamples * (resolution / 8) * channels;
		final AudioFormat format = new AudioFormat(sampleRate, resolution, channels, true, false);

		final Mixer.Info output = (Mixer.Info)outputCombo.getSelectedItem();
		final Mixer.Info input = (Mixer.Info)inputCombo.getSelectedItem();
		
		boolean connected = false;
		contentPane.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));

		try
		{
			connector.connect(input, output, format, bufferSize);
			connected = true;
		}
		catch(LineUnavailableException ex)
		{
			JOptionPane.showMessageDialog(
					contentPane,
					"Unable to open input or output.\nTry again later.",
					"Line unavailable",
					JOptionPane.ERROR_MESSAGE);

			disconnect();
		}
		catch(IllegalArgumentException ex)
		{
			Logger.getLogger(AudioRouterGui.class.getName()).log(Level.WARNING, null, ex);
			
			if(ex.getMessage().contains("format"))
			{
				JOptionPane.showMessageDialog(
						contentPane,
						"Either the input or output does not support the chosen audio format.\nPlease change it and try again.",
						"Unsupported format",
						JOptionPane.ERROR_MESSAGE);
			}

			if(ex.getMessage().contains("buffer"))
			{
				JOptionPane.showMessageDialog(
						contentPane,
						"Buffer size is not a multiple of frame size.\nChange the buffer size and try again.",
						"Invalid buffer size",
						JOptionPane.ERROR_MESSAGE);
			}

			disconnect();
		}
		catch(Exception ex)
		{
			Logger.getLogger(AudioRouterGui.class.getName()).log(Level.WARNING, null, ex);
			
			JOptionPane.showMessageDialog(
					contentPane,
					"An unspecified error occured.\nTry again later.",
					"Exception",
					JOptionPane.ERROR_MESSAGE);

			disconnect();
		}
		
		contentPane.setCursor(Cursor.getDefaultCursor());

		return connected;
	}

	/**
	 * Disconnects input from output.
	 */
	private void disconnect()
	{
		contentPane.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
		connector.disconnect();
		contentPane.setCursor(Cursor.getDefaultCursor());
	}

	/**
	 * Save current settings.
	 */
	private void saveProperties()
	{
		final Properties properties = Properties.getAppProperties();
		
		properties.setProperty(float.class, ((Integer)sampleCombo.getSelectedItem()).floatValue(), "format", "sampleRate");
		properties.setProperty(Integer.class, (Integer)resolutionCombo.getSelectedItem(), "format", "resolution");
		properties.setProperty(Integer.class, (Integer)channelsSpinner.getValue(), "format", "channels");
		
		properties.setProperty(Integer.class, (Integer)bufferCombo.getSelectedItem(), "bufferSize");
		
		properties.setProperty(String.class, ((Mixer.Info)inputCombo.getSelectedItem()).getName(), "mixers", "input");
		properties.setProperty(String.class, ((Mixer.Info)outputCombo.getSelectedItem()).getName(), "mixers", "output");
		
		properties.saveProperties();
	}
}
